#!node
// serverbase.js : v0.5 : Node HTTPS Server
// Author: Eric Flumerfelt, FNAL RSI
// Last Modified: June 3, 2015
// Modified By: Eric Flumerfelt
//
// serverbase sets up a basic HTTPS server and directs requests
// to one of its submodules.
//
// Implementation Notes: modules should assign their emitter to the module_holder[<modulename>] object
// modules will emit 'data' and 'end' signals and implement the function MasterInitFunction()

var cluster = require('cluster');
var fs = require('fs');
var path_module = require('path');
var numCPUs = require('os').cpus().length;
var child_process = require('child_process');
var systemdSocket = require('systemd-socket');

var util = require('util');

var getversion = function () {
    //console.log("Getting Server Version");
    if (fs.existsSync("./version.txt")) {
        //console.log("Reading Server Version from File");
        return "" + fs.readFileSync("./version.txt");
    }
    else {
        child_process.exec("git describe --tags", function (error, stdout, stderr) {
            version = stdout.trim() + "-Git";
            child_process.exec("git status --porcelain", function (error, stdout) {
                if (stdout.length > 0) {
                    version += "*";
                }
            });
        });
    }
}
var version = getversion();

var config = {
    listenhost: "localhost",
    xdaqmode: true,
    destinations: [{name: "main",
		    listenport: 8080,
		    host: "localhost",
		    port: 2015}],
    logFileName: "/tmp/xdaqproxy." + process.env["USER"] + ".log",
    certInfoFile: "/tmp/xdaqproxy." + process.env["USER"] + ".certdata",
    logConsoleEnabled: true
};
function loadConfig() {
    if (fs.existsSync("xdaq_config.json")) {
        config = JSON.parse(fs.readFileSync("xdaq_config.json"));
    } else {
        fs.writeFileSync("xdaq_config.json", JSON.stringify(config));
    }

    if (config.hostname === "localhost" && cluster.isMaster) {
        console.log("Listening only on localhost. To listen on a different address, set \"hostname\" in config.json.\nUse \"0.0.0.0\" to listen on all interfaces.");
    }
}

loadConfig();
var log_file = fs.createWriteStream(config.logFileName, { flags: 'a' });
var log_stdout = process.stdout;

console.log = function (d) { //
    log_file.write((new Date().toString()) + ": " + util.format(d) + '\n');
    if (config.logConsoleEnabled) log_stdout.write(util.format(d) + '\n');
};

function Download(url, path, callback) {
    child_process.execSync("wget --unlink " + url, { cwd: path });
}

function LoadCerts(path) {
    if (!fs.existsSync(path)) {
        console.log("Creating certificates directory");
        fs.mkdirSync(path);

        Download("https://cilogon.org/cilogon-basic.pem", path, function (error) {
            if (error) console.log(error);
        });
        Download("https://cilogon.org/cilogon-basic.crt", path, function (error) {
            if (error) console.log(error);
        });
    }
    var output = [];
    var files = fs.readdirSync(path);
    for (var i = 0; i < files.length; i++) {
        if (files[i].search(".pem") > 0 || files[i].search(".crt") > 0) {
            output.push(fs.readFileSync(path + "/" + files[i]));
        }
    }
    return output;
}

function GetCILogonCRL(path) {
    // Always fetch the latest CRL lists from CILogon:
    // Download("http://crl-cilogon.ncsa-security.net/cilogon-basic.r0", path, function(error) {
    //     if(error) console.log(error);
    //  });
    //        Download("http://crl-cilogon.ncsa-security.net/cilogon-basic.crl", path, function(error) {
    //            if(error) console.log(error);
    //        });
}

function LoadCRLs(path) {
    if (!fs.existsSync(path)) {
        console.log("Creating directory " + path);
        fs.mkdirSync(path);
    }
    GetCILogonCRL(path);
    var output = [];
    var files = fs.readdirSync(path);
    for (var i = 0; i < files.length; i++) {
        if (files[i].search(".r0") > 0 || files[i].search(".crl") > 0) {
            output.push(fs.readFileSync(path + "/" + files[i]));
        }
    }
    return output;
}

function GetDestination(port) {
    var destination = {
        host: "",
        port: 0
    };

    for (var dest in config.destinations) {
	if(config.destinations.hasOwnProperty(dest)) {
            if (config.destinations[dest].listenport === port) {
		destination.host = config.destinations[dest].host;
		destination.port = config.destinations[dest].port;
	    }
        }
    }
    console.log("Destination is http://" + destination.host + ":" + destination.port);
    return destination;
}

// Node.js by default is single-threaded. Start multiple servers sharing
// the same port so that an error doesn't bring the whole system down
if (cluster.isMaster) {

    var destinationWorkers = {};
    function GetMinWorkerCount() {
	var min = 1000000;
	var min_dest = 0;
	for(var dest in destinationWorkers) {
	    if(destinationWorkers.hasOwnProperty(dest)) {
		console.log("Destination " + dest + " has " + destinationWorkers[dest].length + " workers, minimum " + min + " on destination " + min_dest);
		if(destinationWorkers[dest].length < min) {
		    min = destinationWorkers[dest].length;
		    min_dest = dest;
		}
	    }
	}

	console.log("Returning destination "+ min_dest);
	return min_dest;
    }

    for(var dest in config.destinations) {
	if(config.destinations.hasOwnProperty(dest)) {
	    console.log("Creating empty array for destination " + dest);
	    destinationWorkers[dest] = [];
	}
    }

    function StartWorker() {
	var dest = GetMinWorkerCount();
	var env = {};
	env["DESTINATION"] = dest;
        var worker = cluster.fork(env);
	console.log("Worker " + worker.id + " will listen to destination " + dest);
	destinationWorkers[dest].push(worker.id);
    }

    // Start workers for each CPU on the host
    for (var i = 0; i < numCPUs; i++) {
        //for (var i = 0; i < 1; i++) {
	StartWorker();
    }

    // If one dies, start a new one!
    cluster.on("exit", function (worker) {
	var found = false;
	for(var dest in destinationWorkers) {
	    if(destinationWorkers.hasOwnProperty(dest)) {
		for(var i = 0; i < destinationWorkers[dest].length; i++) {
		    if(destinationWorkers[dest][i] == worker.id) {
			found = true;
			destinationWorkers[dest].splice(i,1);
			break;
		    }
		}
	    }
	    if(found){ break; }
	}
	StartWorker();
    });
} else {
    console.log(`Starting worker ${cluster.worker.id}`);
    // Node.js framework "includes"
    var https = require('https');
    var http = require('http');
    var url = require('url');

    console.log("Setting up options");
    var options = {
        key: fs.readFileSync('./certs/server.key'),
        cert: fs.readFileSync('./certs/server.crt'),
        ca: LoadCerts("./certs/certificates"),
        crl: LoadCRLs("./certs/certificates"),
        requestCert: true,
        rejectUnauthorized: false,
        ciphers: "ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-AES256-GCM-SHA384:DHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-SHA256:DHE-RSA-AES128-SHA256:ECDHE-RSA-AES256-SHA384:DHE-RSA-AES256-SHA384:ECDHE-RSA-AES256-SHA256:DHE-RSA-AES256-SHA256:HIGH:!aNULL:!eNULL:!EXPORT:!DES:!RC4:!MD5:!PSK:!SRP:!CAMELLIA:!3DES",
        secureProtocol: "TLSv1_2_method"
    };
    var authlist = " " + fs.readFileSync("./certs/authorized_users");
    console.log("Done setting up options");

    // Make an http server
    var server = https.createServer(options, function (req, res) {
        try {
            console.log("Getting client certificate");
            var clientCertificate = req.connection.getPeerCertificate();
            var username = "";
            var useremail = "";
            var fp = "";
	    var localport = req.connection.localPort;
	    console.log("Local port is " + localport);
            var destination = GetDestination(localport);
            //console.log(JSON.stringify(clientCertificate));

            if (clientCertificate && clientCertificate.subject) {
                console.log("Getting certificate info...");
                username = clientCertificate.subject.CN[0];
                useremail = clientCertificate.subjectaltname.substr(6);
                fp = clientCertificate.modulus;

                console.log("Writing certificate info to temp file");
                fs.writeFileSync(config.certInfoFile, useremail + "\n" + fp);
                setTimeout(function () {
                    if (fs.existsSync(config.certInfoFile) && (Date.now() - fs.statSync(config.certInfoFile).mtime) > 1000) {
                        fs.unlinkSync(config.certInfoFile);
                    }
                }, 1000);
            }

            console.log("Checking if client is authorized");
            console.log("User: " + username + ", Email: " + useremail);
            //console.log(JSON.stringify(req.connection));
            if (req.client.authorized) {
                //console.log(JSON.stringify(clientCertificate));
                //var org = clientCertificate.subject.O[0];
                //if (org !== "Fermi National Accelerator Laboratory") {
                //    readOnly = true;
                //}
                if (authlist.search(username) > 0 || authlist.search(useremail) > 0) {
                    console.log("User: " + username + " (" + useremail + ")");
                }
            }

            console.log("Parsing request url: " + req.url);
            if (req.url.search(/lid=\d+$/) > 0) {
                req.url = req.url + "/";
            }

            var thisurl = url.parse(req.url, true);
            //console.log("Request Querystring before: " + JSON.stringify(thisurl.query));
            thisurl.search = "";

            console.log("Request path: " + req.url);
            if (fp.length > 0 && req.url.search("RequestType=cert") > 0) {
                thisurl.query.httpsUser = fp;
            }
            //console.log("Request Querystring after: " + JSON.stringify(thisurl.query));
            //console.log(JSON.stringify(thisurl));
            var pathname = url.format(thisurl);
            console.log("Adjusted path: " + pathname);

            if (config.xdaqmode && (pathname === "/" || req.url === "/")) {
                var code = 302;
                var redirUrl = "https://" + config.listenhost + ":" + localport + "/urn:xdaq-application:lid=200/";

                console.log("Redirecting to " + redirUrl);
                res.writeHead(code, { 'location': redirUrl });
                res.end();
            }
            else {
                var xdaqpath = pathname;
                var reqOptions = {
                    host: destination.host,
                    port: destination.port,
                    method: req.method,
                    headers: req.headers,
                    path: xdaqpath
                };
                console.log("Request options: " + JSON.stringify(reqOptions));
                var xreq = http.request(reqOptions, function (xres) {
                    if (xres.statusCode >= 300 && xres.statusCode < 400 && xres.headers.location) {
                        console.log("Redirect detected. Going to " + xres.headers.location);
                        var redirUri = url.parse(xres.headers.location + "/");
                        redirUri.hostname = config.listenhost;
                        redirUri.port = localport;
                        redirUri.host = config.listenhost + ":" + localport;
                        redirUri.protocol = "https:";
                        var redirUrl = "/" + url.format(redirUri);
                        console.log("Redirect url adjusted to " + redirUrl);
                        res.writeHead(xres.statusCode, { 'location': redirUrl });
                        res.end();
                    } else {
                        //console.log("Piping xres into res");
                        xres.pipe(res);
                    }
                });
                //console.log("Piping req into xreq");
                req.pipe(xreq);
            }
        } catch (e) {
            console.log("Unhandled error in server: " + JSON.stringify(e));
            console.trace("Unhandled error in server: " + JSON.stringify(e));
        }
    });

    try {
	var dest = process.env["DESTINATION"];
	console.log("Destination index " + dest);

        if (dest == 0) {
            console.log("Listening on https://" + config.listenhost + ":" + config.destinations[dest].port + ", forwarding to http://" + config.destinations[dest].host + ":" + config.destinations[dest].port + " (systemdSocket)");
            server.listen(systemdSocket() || config.destinations[dest].port, config.listenhost);
        } else {
	    if(config.destinations.hasOwnProperty(dest)) {
		console.log("XDAQ Instance " + config.destinations[dest].name + " at http://" + config.destinations[dest].host + ":" + config.destinations[dest].port + " listening on https://" + config.listenhost + ":" + config.destinations[dest].listenport);
		server.listen(config.destinations[dest].listenport, config.listenhost);
	    }
        }
    } catch (e) {
        console.log("Unhandled error in server: " + JSON.stringify(e));
        console.trace("Unhandled error in server: " + JSON.stringify(e));
    }
}
